#include "vmm3a.h"

// Define a map to store the valid range for each feature
std::map<std::string, std::pair<int, int>> featureRanges = {
  {"sm5_sm0", {0, 67}},
  {"sfam", {0, 1}},
  {"st", {0, 3}},
  {"sg", {0, 7}},
  {"stc", {0, 3}},
  {"sdt", {0, 1023}},
  {"sdp10", {0, 1023}},
  {"sc10b", {0, 3}},
  {"sc8b", {0, 3}},
  {"truncate_i", {0, 63}},
  {"nskip_i", {0, 127}},
  {"window_i", {0, 7}},
  {"rollover_i", {0, 4095}},
  {"L0offset_i", {0, 4095}},
  {"offset_i", {0, 4095}}
};

std::map<std::string, std::pair<int, int>> featureChannelRanges = {
  {"sd", {0, 31}},
  {"sz10b", {0, 31}},
  {"sz08b", {0, 15}},
  {"sz06b", {0, 7}},
  {"sc", {0, 1}},
  {"sl", {0, 1}},
  {"st", {0, 1}},
  {"sth", {0, 1}},
  {"sc8b", {0, 3}},
  {"sm", {0, 1}},
  {"smx", {0, 1}}
};

VMM3a::VMM3a(): vmm3aSettings(std::make_unique<VMM3aSettings>()) {
  LoadDefault();
}

void VMM3a::LoadDefault() {
  //VMM3a channel registers
  for(int ch=0; ch<CHANNELS_PER_VMM; ch++){
    vmm3aSettings->channels[ch] = {{"sc", 0}, {"sl", 0}, {"st", 0}, {"sth", 0}, {"sm", 0}, {"sd", 0}, {"smx", 0}, {"sz10b", 0}, {"sz08b", 0}, {"sz06b", 0}  };
  }

  //Fill default values to map for Global Register
  std::vector<std::string> reg_names = {
    "sp", "sdp", "sbmx", "sbft", "sbfp",
    "sbfm", "slg", "sm5_sm0", "scmx", "sfa",
    "sfam", "st", "sfm", "sg", "sng",
    "stot", "sttt", "ssh", "stc",
    "sdt", "sdp10", "sc10b", "sc8b", "sc6b",
    "s8b", "s6b", "s10b", "sdcks", "sdcka",
    "sdck6b", "sdrv", "stpp",
    "slvs", "s32", "stcr", "ssart", "srec",
    "stlc", "sbip", "srat", "sfrst", "slvsbc",
    "slvstp", "slvstk", "slvsdt", "slvsart", "slvstki",
    "slvsena", "slvs6b", "sL0enaV", "slh", "slxh",
    "stgc", "reset1", "reset2", "nskipm_i", "sL0cktest",
    "sL0dckinv", "sL0ckinv", "sL0ena", "truncate_i",
    "nskip_i", "window_i", "rollover_i", "L0offset_i", "offset_i"
  };

  for(const std::string& reg: reg_names) {
    vmm3aSettings->globalRegisters.insert(std::pair<std::string, unsigned short>(reg, 0));
  }

  setRegister("sg", 2);//corrsponds to 3 mV/fC
  setRegister("sm5_sm0", 67); //Temperature_sensor
  setRegister("sdp10", 300);
  setRegister("sdt",300);
  setRegister("s10b",1);
  setRegister("s8b",1);
  setRegister("stc",0);
  setRegister("sc8b",3);
  setRegister("sc10b",3);
  setRegister("srat",0);
  setRegister("sbip",true);
  setRegister("sbfm",true);
  setRegister("sdcks",false);
}

const VMM3a::VMM3aSettings& VMM3a::getVMM3aSettings() const {
  return *vmm3aSettings;
}

vmmStatus VMM3a::setRegister(std::string feature, int val) {
  vmmStatus status = vmmSuccess;
  status = updateRegisterMap(feature, val);
  
  if (status == vmmParamNotFound)
    std::cout << "ERROR register " << feature << " does not exist." << std::endl;
  else if (status == vmmBadValue)
    std::cout << "Value " << val << " is not valid to register: " << feature << std::endl;
  return status;
}

vmmStatus VMM3a::setRegister(std::string feature, int val, int ch) {
  vmmStatus status = vmmSuccess;
  status = updateRegisterMap(feature, val, ch);
  
  if (status == vmmParamNotFound)
    std::cout << "ERROR register " << feature << " does not exist." << std::endl;
  else if (status == vmmBadValue)
    std::cout << "Value " << val << " is not valid to register: " << feature << std::endl;
  else if (status == vmmChannelOutRange)
    std::cout << "Channel " << ch << " out of the range (0 - 63)" << std::endl;
  return status;
}

vmmStatus VMM3a::getRegister(std::string feature, unsigned short& result) const {
  vmmStatus status = vmmSuccess;
  if(vmm3aSettings->globalRegisters.find(feature)!=vmm3aSettings->globalRegisters.end())
    result = vmm3aSettings->globalRegisters.at(feature);
  else
    status = vmmParamNotFound;
  return status;
}

vmmStatus VMM3a::getRegister(std::string feature, int ch, unsigned short& result) const {
  vmmStatus status = vmmSuccess;
  if (ch >= 0 && ch < 64) {
    auto it = vmm3aSettings->channels[ch].find(feature);
    if (it != vmm3aSettings->channels[ch].end())
      result = it->second;
    else
      status = vmmParamNotFound;
  }
  else status = vmmChannelOutRange;
  
  return status;
}

vmmStatus VMM3a::updateRegisterMap(std::string feature, int value) {
  vmmStatus status = vmmSuccess;
  auto it = vmm3aSettings->globalRegisters.find(feature);
  if (it != vmm3aSettings->globalRegisters.end()) {
    auto range = featureRanges.find(feature);
    if (range != featureRanges.end()) {
      int minVal = range->second.first;
      int maxVal = range->second.second;

      if (feature == "sm5_sm0") {
        if(value >= 0 && value <= 63) {
          vmm3aSettings->globalRegisters.at("scmx") = 1;
          vmm3aSettings->globalRegisters.at("sm5_sm0") = value;
        } else if (value >= 64 && value <= 67) {
          //Attention: Starts at 1: 1=Pulser_DAC, 2=Threshold_DAC, 3=Bandgap_reference, 4=Temperature_sensor
          vmm3aSettings->globalRegisters.at("scmx") = 0;
          vmm3aSettings->globalRegisters.at("sm5_sm0") = value - 63;
        } else status = vmmBadValue;
      } 
      else {
        if (value >= minVal && value <= maxVal)
          vmm3aSettings->globalRegisters.at(feature) = value;
        else status = vmmBadValue;
      }
    }
    // If not in the featureRanges, the param expects boolean value.
    else {
      if(value >= 0 && value <= 1)
        vmm3aSettings->globalRegisters.at(feature) = value;
      else status = vmmBadValue;
    }
  }
  else { status = vmmParamNotFound; }
  return status;
}

vmmStatus VMM3a::updateRegisterMap(std::string feature, int value, int ch) {
  vmmStatus status = vmmSuccess;

  if(ch >= 0 && ch < 64) {
    auto range = featureChannelRanges.find(feature);
    if (range != featureChannelRanges.end()) {
      int minVal = range->second.first;
      int maxVal = range->second.second;
      if (value >= minVal && value <= maxVal)
        vmm3aSettings->channels[ch][feature] = value;
      else status = vmmBadValue;
    }
    else status = vmmParamNotFound;
  }
  else status = vmmChannelOutRange;
  return status;
}

vmmStatus VMM3a::getInfo(const std::string& feature, std::string& result) {
  vmmStatus status = vmmSuccess;
  if (vmmInfo.find(feature) != vmmInfo.end())
    result = vmmInfo[feature];
  else 
    status = vmmParamNotFound;
  return status;
}

vmmStatus VMM3a::setInfo(const std::string& feature, const std::string& value) {
  vmmStatus status = vmmSuccess;
  if (vmmInfo.find(feature) != vmmInfo.end())
    vmmInfo[feature] = value;
  else 
    status = vmmParamNotFound;
  return status;
}

VMM3a::~VMM3a(){}
